/***
 *  Elliptic Basis for Finite Fields
 *
 *  Distributed under the terms of the GNU Lesser General Public License (L-GPL)
 *                  http://www.gnu.org/licenses/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *  Copyright 2007-2008 R. Lercier & J.-M. Couveignes
 */

/***
 *
 * This file provides MAGMA routines to handle basis for extensions
 * (of degree d) of finite fields (with q elements) such that there
 * exists fast algorithms to compute Frobenius and
 * multiplications. These basis, available when there exists a point
 * of order d on an elliptic curve defined over GF(q), are called
 * "elliptic (normal) basis" in [CoLe2008] and, at some extend, can be
 * seen as a a generalisation of Gaussian Normal Basis. It is worth to
 * add that this code proposes three basis (polynomial, normal and
 * non-normal basis) with arithmetic operations for them, as well as
 * changes of variables with good complexity in d.
 *
 * More precisely, given a MAGMA finite field Fq, with q elements, a
 * call 'EB := EllBasisCreate(Fq, d)' returns a structure 'EB' of type
 * 'EllBasis' such that the field 'EB`Initialised' is set to true if a
 * degree d elliptic extension has been succesfully initialised, and
 * to false otherwise. When everything's ok, the field 'EB`Fqd'
 * contains a MAGMA degree d extension defined by a polynomial
 * basis. Elements in the elliptic polynomial basis are thus given by
 * polynomials in 'Fqd.1'.
 *
 * Calls to 'EllPolynomialToBasis' (or 'EllPolynomialToNormalBasis')
 * sends an element given in the polynomial basis to the elliptic
 * (normal) basis. Elements in these two basis are then given by
 * elements of a vector space of dimension d over Fq (that is elements
 * of 'VectorSpace(Fq, d)' with the MAGMA syntax). Furthermore, calls
 * to 'EllBasisToNormalBasis' or 'EllNormalBasisToBasis' enables to
 * switch between normal and non-normal elliptic basis.
 *
 * In any basis, additions are performed through the '+' MAGMA
 * operation. Routines for multiplications, inverses, Frobenius and
 * exponentiations are provided as well but depend on the basis. In
 * the elliptic polynomial basis, we handle these operations through
 * the usual MAGMA syntax, '*', '1/a', etc. In the elliptic basis, we
 * provide for this task the routines 'EllBasisMultiply',
 * 'EllBasisFrobenius', 'EllBasisFrobeniusInverse', 'EllBasisInverse'
 * and ' EllBasisExp'. Finally, in normal basis, similar routines,
 * denoted 'EllNormalBasisOperation' instead of 'EllBasisOperation',
 * are available.
 *
 *********************************************************************/

/***
 * Bibliography
 *
 * [CoLe2008] "Elliptic Basis for finite fields", J.-M. Couveignes and
 * R. Lercier, To appear in Finite Fields and their Applications, 2008.
 *
 *********************************************************************/

/***
 * Changes.
 *
 * 2008-05-27, v1.1 : R. Lercier
 *              Fast algorithm support added
 *
 * 2007-12-29, v1.0 : R. Lercier
 *              Initial version.
 *
 ********************************************************************/

/***
 * Exported functions
 *
 * intrinsic EllBasisToPolynomial(A::ModTupFldElt,EllBasis::Rec) -> FldFinElt
 * intrinsic EllPolynomialToBasis(a::FldFinElt, EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllNormalBasisToPolynomial(A::ModTupFldElt, EllBasis::Rec) -> FldFinElt
 * intrinsic EllPolynomialToNormalBasis(a::FldFinElt, EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllNormalBasisToBasis(A::ModTupFldElt, EllBasis::Rec) -> FldFinElt
 * intrinsic EllBasisToNormalBasis(A::ModTupFldElt, EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllBasisFrobenius(A::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllNormalBasisFrobenius(A::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllBasisFrobeniusInverse(A::ModTupFldElt,EllBasis) -> ModTupFldElt
 * intrinsic EllNormalBasisFrobeniusInverse(A::ModTupFldElt,EllBasis) -> ModTupFldElt
 * intrinsic EllBasisMultiply(A::ModTupFldElt,B::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllNormalBasisMultiply(A::ModTupFldElt,B::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllBasisInverse(A::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllBasisInverse(A::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllNormalBasisInverse(A::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllBasisExp(A::ModTupFldElt,e::RngIntElt,EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllNormalBasisExp(A::ModTupFldElt,e::RngIntElt,EllBasis::Rec) -> ModTupFldElt
 * intrinsic EllBasisCreate(Fq::FldFin, d::RngIntElt :
 *           CheckIfEllipticBasisExist := true, FastAlgorithmsOnly := false) -> Rec
 *
 ********************************************************************************/

declare verbose EllBasis, 1;

EllBasis := recformat<Initialised, FastAlgorithmsOnly, /* Initialization flag */
    Fq, d, Fqd, Ec, Xi, Yi,               /* Common data */
    B, MB, Xs, Xs0, XsS,                  /* Elliptic basis data */
    NB, MNB, csta, cstb,                  /* Normal Elliptic basis data */
    FastAlgorithmsAvailable, Ur, Uri, Xr, Iota,         /* Normal Fast Multiplication data*/
    NaiveConversion, TD, LD, dn, yP, py1, py2, pi1, pi2, xi, yb  /* Fast conversion data */
    >;                   

/***
 * Internal functions
 *********************/

/* The slope between two points on  an elliptic curve*/
EllGamma := function(EllBasis, k, l)

  a1, a2, a3, a4, a6 := Explode(aInvariants(EllBasis`Ec));
  d := EllBasis`d;;

  if k eq l or k eq d-l then
      return -(3*EllBasis`Xi[1+k]^2+a1*(EllBasis`Yi[1+k]+a1*EllBasis`Xi[1+k]+a3)+
	  2*a2*EllBasis`Xi[1+k]+a4)/(2*EllBasis`Yi[1+k]+a1*EllBasis`Xi[1+k]+a3);
  end if;
 
  return (EllBasis`Yi[1+l]+EllBasis`Yi[1+k]+a1*EllBasis`Xi[1+k]+a3) /
      (EllBasis`Xi[1+l]-EllBasis`Xi[1+k]);
end function;

/* Looking for an elliptic curve of prescribed order */
EllFindEcOfPrescribedOrder := function(Fp, order)

    found := false;
    repeat
	for i := 1 to 10 do
	    repeat
		repeat
		    a1 := Random(Fp); a2 := Random(Fp); a3 := Random(Fp);
		    a4 := Random(Fp); a6 := Random(Fp);

		    Disc := -a1^6*a6 + a1^5*a3*a4 - a1^4*a2*a3^2 - 12*a1^4*a2*a6 + a1^4*a4^2 + 
			8*a1^3*a2*a3*a4 + a1^3*a3^3 + 36*a1^3*a3*a6 - 8*a1^2*a2^2*a3^2 - 
			48*a1^2*a2^2*a6 + 8*a1^2*a2*a4^2 - 30*a1^2*a3^2*a4 + 72*a1^2*a4*a6 + 
			16*a1*a2^2*a3*a4 + 36*a1*a2*a3^3 + 144*a1*a2*a3*a6 - 96*a1*a3*a4^2 - 
			16*a2^3*a3^2 - 64*a2^3*a6 + 16*a2^2*a4^2 + 72*a2*a3^2*a4 + 288*a2*a4*a6 - 
			27*a3^4 - 216*a3^2*a6 - 64*a4^3 - 432*a6^2;

		until Disc ne 0;
		Ec := EllipticCurve([a1, a2, a3, a4, a6]);
	    until (#Ec mod order) eq 0;

	    P := Random(Ec);
	    if (Order(P) mod order) eq 0 and  (Order(P) mod order^2) ne 0 then
		found := true; break;
	    end if;
	end for;
    until found eq true;

    return Ec, (Order(P) div order)*P;

end function;

/* Checking if there exists an elliptic curve with the given order */
EllDoesItExistOneECWithThisOrder := function(Fq, d)
    q := #Fq; p:= Characteristic(Fq); n := Degree(Fq); m := q+1-d;

    /* d can't be too small nor too large */
    if m^2 gt 4*q then return false; end if;

    /* When p divides m, only the (very few) supersingular curves exists  */
    if (m mod p) eq 0 then
	if IsOdd(n) then
	    if m ne 0 and
	       m ne Floor(Sqrt(p*q)) and
	       m ne -Floor(Sqrt(p*q)) then
	       return false;
	    end if;
	else
	    if (m ne 2*Floor(Sqrt(q)) and
	        m ne -2*Floor(Sqrt(q))) and
	       ((m ne Floor(Sqrt(q)) and
	         m ne -Floor(Sqrt(q))) or
		(p ne 3 and (p mod 3) ne 2)) and
	       ((m ne 0) or
	       (((p mod 4) ne 2) and ((p mod 4) ne 3))) then
	       return false;
	    end if;
	end if;
    end if;

    return true;
    
end function;

/* Checking if there exists an elliptic basis of degre d over Fq */
EllIsPossible := function(Fq, d)

    q := #Fq; p:= Characteristic(Fq); n := Degree(Fq); m := q+1-d;

    if d le 1 then return false; end if;
    if d eq q-1 and not q in [2, 2^2, 2^3, 3, 5, 7] then return false; end if;

    k := Ceiling(q+1-2*Sqrt(q)) div d;
    dp := k*d; found := false;
    while dp le Floor(q+1+2*Sqrt(q)) and found eq false do
	found := EllDoesItExistOneECWithThisOrder(Fq, dp);
	dp +:= d;
    end while;

    return found;
end function;

/* The function U_OA evaluated at P */
EllUOA := function(A, P, Ec)

    if A eq Ec!0 then return 1; end if;

    a1, a2, a3, a4, a6 := Explode(aInvariants(Ec));
    x1, y1, _ := Explode(Eltseq(A));
    x, y,   _ := Explode(Eltseq(P));

    return (y+y1+a1*x1+a3)/(x-x1);
end function;

/* The function U_AB evaluated at P */
EllUAB := function(A, B, P, Ec)

    a1, a2, a3, a4, a6 := Explode(aInvariants(Ec));

    if A eq Ec!0 then
	return EllUOA(B, P, Ec);
    end if;

    if B eq Ec!0 then
	return -EllUOA(A, P, Ec)-a1;
    end if;

    x1, y1, _ := Explode(Eltseq(A));
    x2, y2, _ := Explode(Eltseq(B));
    x,  y,  _ := Explode(Eltseq(P));

    if A eq -B then
	return -(3*x1^2 + 2*a2*x1 + a4 -a1*y1) / (2*y1+a1*x1+a3)-(x*a1+a3+2*y1)/(x-x1);
    end if;
 
    return (y1+y2+a3+a1*x1)/(x2-x1)-((-x2+x1)*y+(-y2+y1+a1*(x1-x2))*x+(-(y1+a1*x1+a3)*x2+x1*(y2+a1*x2+a3)))/(x-x1)/(x-x2);

end function;

/* A naive convolutional product

EllConvolutionalProduct := function(A, B)
    C := 0*A; d := Ncols(A);
    for j := 0 to d-1 do
	for i := 0 to d-1 do
	    C[1+j] +:= A[1+i]*B[1+((j-i) mod d)];
	end for;
    end for;
    return C;
end function;

*/

/* Convolutional product

  We take adavantage of Magma's implementation of polynomial products
  to get a not too slow convolution product.
*/
EllConvolutionalProduct := function(A, B)

  Fq := CoefficientRing(A); d := Ncols(A);
  Px := PolynomialRing(CoefficientRing(A));
  pC := Px!Eltseq(A) * Px!Eltseq(B);
  Cp := VectorSpace(Fq, d)!0;
  for i := 0 to d-1 do
    Cp[1+i] := Coefficient(pC, i);
  end for;
  for i := d to 2*d-1 do
    Cp[1+i-d] +:= Coefficient(pC, i);
  end for;

  return Cp;
end function;

/* Dot product */
EllDotProduct := function(A, B)
    C := 0*A; d := Ncols(A);
    for j := 0 to d-1 do
	C[1+j] +:= A[1+j]*B[1+j];
    end for;
    return C;
end function;

/* Circular shift */
EllSigma := function(A)
    C := 0*A; d := Ncols(A);
    for j := 0 to d-1 do
	C[1+j] +:= A[1+((j-1) mod d)];
    end for;
    return C;
end function;

/* Fast Multiplication for Elliptic Normal Basis */
EllNormalBasisFastMultiply := function(A, B, EllBasis)

    UrA := EllConvolutionalProduct(EllBasis`Ur, A);
    UrB := EllConvolutionalProduct(EllBasis`Ur, B);

    V :=  EllConvolutionalProduct(EllBasis`Xr,
	EllDotProduct(A-EllSigma(A), B-EllSigma(B)));

    V1 := EllConvolutionalProduct(EllBasis`Uri,
	-V + EllDotProduct(UrA, UrB));

    C :=  V1+EllConvolutionalProduct(
	EllBasis`Iota, EllDotProduct(A-EllSigma(A), B-EllSigma(B)));

    return C;
end function;

/* Computing a tree with polynomials TT(x-xi),
   in an efficient way */
forward EllEvaluationLevel;
procedure EllEvaluationLevel(~L, Xi, bg, ed)
    df := ed-bg;
    if df gt 2 then
      dlt := ((ed-bg + 1) div 2);
      EllEvaluationLevel(~L, Xi, bg, bg + dlt);
      EllEvaluationLevel(~L, Xi, bg + dlt, ed);
    elif df eq 2 then
      Append(~L, Xi[bg]*Xi[bg+1]);  
    elif df eq 1 then
      Append(~L, Xi[bg]);
    end if;
end procedure;
procedure EllEvaluationTree(~L)
    while #L[#L] gt 1 do
        lst := [];
        EllEvaluationLevel(~lst, L[#L], 1, #L[#L]+1);
	Append(~L, lst);
    end while;
end procedure;

/* Computing efficiently U(x) and V(x), such that
      (U(x) + y * V(x)) / TT (x - xi) = sum_i A[i] * (y + ci) / (x -xi)
*/
forward EllRationalFraction;
procedure EllRationalFraction(~NewNumY, ~NewNumC,
    NumY, NumC, Den, bg, ed,
    EllBasis)

    df := ed-bg;

    if df gt 2 then

	dlt := ((ed-bg + 1) div 2);

	EllRationalFraction(~NewNumY, ~NewNumC,
	    NumY, NumC, Den, bg, bg+dlt,
	    EllBasis);

	EllRationalFraction(~NewNumY, ~NewNumC,
	    NumY, NumC, Den, bg+dlt, ed,
	    EllBasis);

    elif df eq 2 then

	Append(~NewNumY, NumY[bg]*Den[bg+1] + NumY[bg+1]*Den[bg]);
	Append(~NewNumC, NumC[bg]*Den[bg+1] + NumC[bg+1]*Den[bg]);

    elif df eq 1 then

	Append(~NewNumY, NumY[bg]);
	Append(~NewNumC, NumC[bg]);

    end if;

end procedure;

/* Evaluating a polynomial B at many points xi */
forward EllFastEvaluate;
EllFastEvaluate := procedure(~NewLB, ~ptr, LB, TD, bg, ed)

    df := ed-bg;

    if df gt 2 then
	
	dlt := ((ed-bg + 1) div 2);

	EllFastEvaluate(~NewLB, ~ptr, LB, TD, bg, bg+dlt);
	EllFastEvaluate(~NewLB, ~ptr, LB, TD, bg+dlt, ed);

    elif df eq 2 then

	Append(~NewLB, LB[ptr] mod TD[bg]);
	Append(~NewLB, LB[ptr] mod TD[bg+1]);
	ptr +:= 1;

    elif df eq 1 then

	Append(~NewLB, LB[ptr] mod TD[bg]);
	ptr +:= 1;

    end if;

end procedure;

/* Half gcd to get y1, y2 s.t y(x) = y1(x) / y2(x) mod b(x)
   with y1 of degree bounded by deg */
EllConvergent := function(_a, _b, deg)

  a := _a; b := _b;
  x := 0; lastx := 1;
  y := 1; lasty := 0;

  while Degree(b) gt deg do
    tmp := b;
    q := a div b; b := a - b*q;
    a := tmp;

    tmp := x; x := lastx - q*x; lastx := tmp;
    tmp := y; y := lasty - q*y; lasty := tmp;
  
  end while;
  
  return b, y;
end function;

/***
 * Exported functions
 *********************/

/* Elliptic Basis To Polynomial Basis Conversion

   Programmed in a ``quasi-linear in d complexity'' style.
*/
intrinsic EllBasisToPolynomial(A::ModTupFldElt,EllBasis::Rec) -> FldFinElt
    {Convert in a naive way an element given in the elliptic basis to an element in the elliptic extension}

    d := EllBasis`d;
    
    if EllBasis`NaiveConversion then
	return &+[A[i]* EllBasis`B[i] : i in [1..d]];
    end if;

    a1, a2, a3, a4, a6 := Explode(aInvariants(EllBasis`Ec));
    x := PolynomialRing(EllBasis`Fq).1;

    NumY := []; NumC := [];
    for i := 1 to ((d-1) div 2) do
	Append(~NumY, 0*x+A[1+i] + A[1+d-i]);
	Append(~NumC, 0*x+
	    A[1+i]  * ( EllBasis`Yi[1+i]   + a1 * EllBasis`Xi[1+i]    + a3 ) +
	    A[1+d-i]* ( EllBasis`Yi[1+d-i] + a1 * EllBasis`Xi[1+d-i]  + a3 ));
    end for;
    if IsEven(d) then
	dlt := (d div 2);
	Append(~NumY, A[1+dlt]);
	Append(~NumC,  A[1+dlt] * (EllBasis`Yi[1+dlt]+a1*EllBasis`Xi[1+dlt]+a3));
    end if;

    for lvl := 1 to #EllBasis`TD-1 do
	NewNumY := []; NewNumC := [];
	EllRationalFraction(~NewNumY, ~NewNumC,
	    NumY, NumC, EllBasis`TD[lvl], 1, #EllBasis`TD[lvl]+1,
	    EllBasis);
	NumY := NewNumY; NumC := NewNumC;
    end for;

    ny := EllBasis`Fqd!Eltseq(EllBasis`Fqd!(NumY[1]), EllBasis`Fq);
    nc := EllBasis`Fqd!Eltseq(EllBasis`Fqd!(NumC[1]), EllBasis`Fq);

    return ( (ny * EllBasis`yP + nc) / EllBasis`dn  + A[1+0] );

end intrinsic;

/* Polynomial Basis to Elliptic Basis Conversion

   Programmed in a ``quasi-linear complexity in d'' style.
*/
intrinsic EllPolynomialToBasis(a::FldFinElt, EllBasis::Rec) -> ModTupFldElt
    {Convert an element given in the elliptic extension to the elliptic basis}

    Fq := EllBasis`Fq; d := EllBasis`d; Fqd := EllBasis`Fqd;
    if EllBasis`NaiveConversion then
	return (VectorSpace(Fq, d)!Eltseq(a, Fq)) * EllBasis`MB;
    end if;

    A := VectorSpace(Fq, d)!0;
    
    a1, a2, a3, a4, a6 := Explode(aInvariants(EllBasis`Ec));
    Px := PolynomialRing(EllBasis`Fq);

    pD := Px!Eltseq(EllBasis`dn, Fq);
    py1 := EllBasis`py1; py2 := EllBasis`py2;
    pi1 := EllBasis`pi1; pi2 := EllBasis`pi2;
    LD := EllBasis`LD;

    /* U and V s.t a = alpha_0 + (U + y V) / D  */
    pt := Px!Eltseq(EllBasis`dn*a*Fqd!Eltseq(py2), Fq);
    pV := ((pt mod py2) * pi1 ) mod py2;
    if IsOdd(d) then
	pU := (pt - pV*py1) div py2;
    else
	pU := ((pt mod py1) * pi2 ) mod py1;

	sk :=
	    Coefficient(pt-pU*py2-pV*py1, Valuation(py1)+Valuation(py2)) /
	    (Coefficient(py1,Valuation(py1))*Coefficient(py2,Valuation(py2)));

	xi := EllBasis`Xi[1+(d div 2)]; yi := EllBasis`Yi[1+(d div 2)];

	k1 :=
	    (Evaluate(pU, xi)+yi*(Evaluate(pV, xi)+sk*Evaluate(py2, xi))) /
	    (-Evaluate(py1, xi)+yi*Evaluate(py2, xi));
	k2 := sk-k1;

	pV +:= k2 * py2;
	pU +:= k1 * py1;
    end if;
    if Degree(pU) eq (d div 2) then
	A[1+0] := Coefficient(pU, d div 2);
    else
	A[1+0] := 0;
    end if;
    pU -:= A[1+0]*Px!Eltseq(EllBasis`dn, Fq);
    
    /* Evaluation of pU at Xi's */
    lvl := #EllBasis`TD;
    LU := [pU mod EllBasis`TD[lvl][1]];

    for lvl := #EllBasis`TD-1 to 1 by -1 do
	NewLU := []; ptr := 1;
	EllFastEvaluate(~NewLU, ~ptr,
	    LU, EllBasis`TD[lvl], 1, #EllBasis`TD[lvl]+1);
	LU := NewLU;
    end for;

    /* Evaluation of pV at Xi's */
    lvl := #EllBasis`TD;
    LV := [pV mod EllBasis`TD[lvl][1]];

    for lvl := #EllBasis`TD-1 to 1 by -1 do
	NewLV := []; ptr := 1;
	EllFastEvaluate(~NewLV, ~ptr,
	    LV, EllBasis`TD[lvl], 1, #EllBasis`TD[lvl]+1);
	LV := NewLV;
    end for;

    /* Getting A_i's */
    for i := 1 to ((d-1) div 2) do
	b := Coefficient(LU[i], 0) + EllBasis`Yi[1+i] * Coefficient(LV[i], 0);
	bi := Coefficient(LU[i], 0) + EllBasis`Yi[1+d-i] * Coefficient(LV[i], 0);

	c  := EllBasis`Yi[1+i] + a1 * EllBasis`Xi[1+i] + a3;
	ci := EllBasis`Yi[1+d-i] + a1 * EllBasis`Xi[1+d-i] + a3;

	A[1+i]   :=   b/(c-ci)/LD[i];
	A[1+d-i] :=  -bi/(c-ci)/LD[i];
    end for;

    /* Getting A_(d/2) */
    if IsEven(d) then

	xi := EllBasis`xi; yb := EllBasis`yb;
	
	tmp :=
	    (Evaluate(pU, xi) + yb *  Evaluate(pV, xi)) / Evaluate(pD,xi);
	    
	for i := 1 to ((d-1) div 2) do

	    tmp -:= A[1+i] *
		(yb + EllBasis`Yi[1+i] + a1 * EllBasis`Xi[1+i] + a3) /
		(xi - EllBasis`Xi[1+i]);
	    tmp -:= A[1+d-i] *
		(yb + EllBasis`Yi[1+d-i] + a1 * EllBasis`Xi[1+d-i] + a3) /
		(xi - EllBasis`Xi[1+d-i]);

	end for;

	A[1+(d div 2)] := tmp * (xi - EllBasis`Xi[1+(d div 2)]) /
	    (yb + EllBasis`Yi[1+(d div 2)] + a1 * EllBasis`Xi[1+(d div 2)] + a3);

    end if;

    return A;
end intrinsic;

/* Elliptic Normal Basis To Polynomial Basis Conversion
*/
intrinsic EllNormalBasisToPolynomial(A::ModTupFldElt, EllBasis::Rec) -> FldFinElt
    {Convert a element given in the elliptic normal basis to an element in the elliptic extension}

    d := EllBasis`d;
    if EllBasis`NaiveConversion then
	return &+[A[i]* EllBasis`NB[i] : i in [1..d]];
    end if;

    B := EllNormalBasisToBasis(A, EllBasis);
    a := EllBasisToPolynomial(B, EllBasis);

    return a;
end intrinsic;

/* Elliptic Polynomial Basis To Normal Basis Conversion
*/
intrinsic EllPolynomialToNormalBasis(a::FldFinElt, EllBasis::Rec) -> ModTupFldElt
    {Convert an element given in the elliptic extension to the
    elliptic normal basis}

    Fq := EllBasis`Fq; d := EllBasis`d;

    if EllBasis`NaiveConversion then
	return (VectorSpace(Fq, d)!Eltseq(a, Fq)) * EllBasis`MNB;
    end if;

    B := EllPolynomialToBasis(a, EllBasis);
    A := EllBasisToNormalBasis(B, EllBasis);

    return A;
end intrinsic;

/* Elliptic Normal Basis To Elliptic Basis Conversion

   Code with linear complexity in d.
*/
intrinsic EllNormalBasisToBasis(A::ModTupFldElt, EllBasis::Rec) -> FldFinElt
    {Convert a element given in the elliptic normal basis to an element in the elliptic basis}

    d := EllBasis`d;
    a := EllBasis`csta; b := EllBasis`cstb;
    a1, _, _, _, _ := Explode(aInvariants(EllBasis`Ec));

    B := Parent(A)!0;

    /* k = 0 */
    B[1+0] +:= b * A[1+0];
    B[1+1] +:= a * A[1+0];

    /* k != 0, d-1 */
    for k := 1 to d-2 do
	B[1+0]   +:=  (b + a * EllGamma(EllBasis, k, k+1)) * A[1+k];
	B[1+k]   +:= - a * A[1+k];
	B[1+k+1] +:=   a * A[1+k];
    end for;

    /* k = d-1 */
    B[1+0]   +:= (b - a1 * a ) * A[1+d-1];
    B[1+d-1] +:= - a * A[1+d-1];

    return B;
end intrinsic;

/* Elliptic Basis To Elliptic Normal Basis Conversion 

   Code with linear complexity in d.
*/
intrinsic EllBasisToNormalBasis(A::ModTupFldElt, EllBasis::Rec) -> ModTupFldElt
    {Convert an element given in the elliptic basis to the
    elliptic normal basis}

    Fq := EllBasis`Fq; d  := EllBasis`d;
    ai := 1/EllBasis`csta; b := EllBasis`cstb;
    a1, _, _, _, _ := Explode(aInvariants(EllBasis`Ec));

    B := Parent(A)!0;

    cf := A[1+0] - b*ai*A[1+1];
    if d gt 2 then
	cf +:= (b*ai - a1)*A[1+d-1];
    end if;
    lbd := Fq!0;
    for k := 2 to d-2 do
	lbd +:= EllGamma(EllBasis, k-1, k);
	cf -:= (k*b*ai+lbd)*A[1+k];
    end for;

    /* Any k */
    for j := 0 to d-1 do B[1+j] := cf; end for;

    /* k = 1 */
    B[1+0] +:= ai * A[1+1];

    /* k != 0, 1, d-1 */
    cf := Fq!0;
    for k := d-2 to 2 by -1 do 
	cf +:= ai * A[1+k];
	B[1+k-1] +:= cf;
    end for;
    B[1+0] +:= cf;

    /* k = d-1 */
    if d gt 2 then
	B[1+d-1] +:= -ai * A[1+d-1];
    end if;
    
    return B;
end intrinsic;

/* Elliptic Basis Frobenius 

   Code with linear complexity in d.
*/
intrinsic EllBasisFrobenius(A::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
    {Frobenius of an element given in the elliptic basis}

    Fq := EllBasis`Fq;
    d := EllBasis`d;
    a1, _, _, _, _ := Explode(aInvariants(EllBasis`Ec));

    C := 0*A;

    C[1+0] := A[1+0] - a1*A[1+1];

    if d gt 2 then
	C[1+0] +:= &+[A[1+k]*EllGamma(EllBasis, d-1, k-1) : k in [2..(d-1)]];
    end if;
    
    for k := 1 to d-2 do C[1+k] := A[1+k+1]; end for;

    C[1+d-1] := -&+[A[1+k] : k in [1..(d-1)]];

    return C;
end intrinsic;


/* Elliptic Normal Basis Frobenius

   Code with linear complexity in d.
*/
intrinsic EllNormalBasisFrobenius(A::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
    {Frobenius of an element given in the elliptic basis}

    Fq := EllBasis`Fq;
    d := EllBasis`d;

    C := Parent(A)!0;

    C[1+d-1] := A[1+0];
    for k := 0 to d-2 do
	C[1+k] := A[1+k+1];
    end for;

    return C;
end intrinsic;

/* Inverse of the Elliptic Basis Frobenius 

   Code with linear complexity in d.
*/
intrinsic EllBasisFrobeniusInverse(A::ModTupFldElt,EllBasis) -> ModTupFldElt
    {Frobenius of an element given in the elliptic basis}

    Fq := EllBasis`Fq;
    d := EllBasis`d;
    a1, _, _, _, _ := Explode(aInvariants(EllBasis`Ec));

    C := 0*A;

    C[1+0] := A[1+0] - a1*A[1+d-1];


    if d gt 2 then
	C[1+0] +:= &+[A[1+k]*EllGamma(EllBasis, k, d-1) : k in [1..(d-2)]];

    end if;

    C[1+1] := -&+[A[1+k] : k in [1..(d-1)]];

    for k := 2 to d-1 do C[1+k] +:= A[1+k-1]; end for;

    return C;
end intrinsic;

/* Inverse of the Elliptic Normal Basis Frobenius

   Code with linear complexity in d.
 */
intrinsic EllNormalBasisFrobeniusInverse(A::ModTupFldElt,EllBasis) -> ModTupFldElt
    {Frobenius of an element given in the elliptic basis}

    Fq := EllBasis`Fq;
    d := EllBasis`d;

    C := Parent(A)!0;

    C[1+0] := A[1+d-1];
    for k := 1 to d-1 do
	C[1+k] := A[1+k-1];
    end for;


    return C;
end intrinsic;

/* Elliptic Basis Multiplication

   By default, quadratic complexity in d
*/
intrinsic EllBasisMultiply(A::ModTupFldElt,B::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
    {Multiplication of two elements given in the elliptic basis}

    if EllBasis`FastAlgorithmsOnly then	
	An := EllBasisToNormalBasis(A, EllBasis);
	Bn := EllBasisToNormalBasis(B, EllBasis);
	Cn := EllNormalBasisFastMultiply(An, Bn, EllBasis);
	return EllNormalBasisToBasis(Cn, EllBasis);
    end if;
    
    Fq := EllBasis`Fq; d := EllBasis`d; Xi := EllBasis`Xi;
    a1, a2, a3, a4, a6 := Explode(aInvariants(EllBasis`Ec));
    Xs := Parent(A)!EllBasis`Xs; XsS := EllBasis`XsS;
    
    C := 0*A;
    
    /* Correcting terms when a1 <> 0 */
    a := 0; b := B[1+1]; C[1+1] -:= a1*A[1+1]*b;

    for k := 2 to d-1 do
	a +:= A[1+k-1]; b +:= B[1+k];
	C[1+k] -:= a1*(A[1+k]*b + B[1+k]*a);
    end for;

    /* x_b */
    a +:= A[d];
    C +:= a*b*Xs; C[1+0] +:= a*b*a2;

    /*  x_k + x_l */
    ap := &+[Eltseq(A)[i]*Xi[i] : i in [2 .. d]];
    bp := &+[Eltseq(B)[i]*Xi[i] : i in [2 .. d]];
    C[1+0] +:= a*bp + ap*b;

    /* Case l = k = 0 */
    c := A[1+0]*B[1+0];
    C[1+0] +:= c;

    /* Case l = 0 or k = 0, l + k <> 0 */
    for k := 1 to d-1 do
	c := A[1+k]*B[1+0] + A[1+0]*B[1+k];
	C[1+k] +:= c;
    end for;
    
    /* Case l = d/3, k = 2*d/3 */
    if d mod 3 eq 0 then
	
	k := 2*(d div 3); l := d div 3;
	gkl := EllGamma(EllBasis, k, l);
	
	ckl := A[1+k]*B[1+l]; clk := A[1+l]*B[1+k];
	C[1+k] -:= gkl*(ckl+clk);
	C[1+l] +:= gkl*(clk+ckl); 
	
    end if;

   /* Case k = 2*l.

      Orbit given by

               (2*l, l), (d-l, d-2*l), (d-l, l)

               and their transposed.

   */
   for k := 2 to Floor((2*d-1)/3) by 2 do
       l := k div 2;

       gkl := EllGamma(EllBasis, k, l);

       i1 := 2*l; i2 :=  d-l; j1 := d-2*l; j2 := l;

       ci1j2 := A[1+i1]*B[1+j2]; cj2i1 := A[1+j2]*B[1+i1];
       ci2j1 := A[1+i2]*B[1+j1]; cj1i2 := A[1+j1]*B[1+i2];
       ci2j2 := A[1+i2]*B[1+j2]; cj2i2 := A[1+j2]*B[1+i2];

       C[1+i1] -:= gkl*(ci1j2 + cj2i1);
       C[1+i2] -:= gkl*(ci2j1 + cj1i2 +
	                ci2j2 + cj2i2);

       C[1+j1] +:= gkl*(cj1i2 + ci2j1);
       C[1+j2] +:= gkl*(cj2i1 + ci1j2 +
	                cj2i2 + ci2j2);

   end for;

   /* Case l = 2*k 

      Orbit given by

               (k,2*k), (2*d-2*k,d-k), (k,d-k)

               and their transposed.

   */
   for k := Floor(d/2+1) to  Floor((2*d-1)/3) do
       l := (2*k) mod d;

       gkl := EllGamma(EllBasis, k, l);



       i1 := k; i2 :=  (2*d - 2*k) mod d; j1 := (2*k) mod d; j2 := d - k;

       ci1j1 := A[1+i1]*B[1+j1]; cj1i1 := A[1+j1]*B[1+i1];
       ci2j2 := A[1+i2]*B[1+j2]; cj2i2 := A[1+j2]*B[1+i2];
       ci1j2 := A[1+i1]*B[1+j2]; cj2i1 := A[1+j2]*B[1+i1];

       C[1+i1] -:= gkl*(ci1j1 + cj1i1 +
	                ci1j2 + cj2i1);
       C[1+i2] -:= gkl*(ci2j2 + cj2i2);

       C[1+j1] +:= gkl*(cj1i1 + ci1j1);
       C[1+j2] +:= gkl*(cj2i2 + ci2j2 +
	                cj2i1 + ci1j2);
   end for;
   
   /* Other cases in the fundamental domain.

      Orbit given by

               (k,l), (d-l,d-k), (k,k-l),
               (d-k+l,d-k), (d-k+l,l), (d-l,k-l)

               and their transposed.

   */
   for k := 3 to Floor((2*d-1)/3)  do
       for l := Max(1, 2*k-d + 1) to Floor((k-1)/2) do
	       gkl := EllGamma(EllBasis, k, l);

         i1 := k; i2 := d-l; i3 := d-k+l;
	 j1 := d-k; j2 := l; j3 := k-l;

	 ci1j2 := A[1+i1]*B[1+j2]; cj2i1 := A[1+j2]*B[1+i1];
	 ci1j3 := A[1+i1]*B[1+j3]; cj3i1 := A[1+j3]*B[1+i1];

	 ci2j1 := A[1+i2]*B[1+j1]; cj1i2 := A[1+j1]*B[1+i2];
	 ci2j3 := A[1+i2]*B[1+j3]; cj3i2 := A[1+j3]*B[1+i2];

	 ci3j1 := A[1+i3]*B[1+j1]; cj1i3 := A[1+j1]*B[1+i3];
	 ci3j2 := A[1+i3]*B[1+j2]; cj2i3 := A[1+j2]*B[1+i3];

	 C[1+i1] -:= gkl*(ci1j2 + cj2i1 +
	                  ci1j3 + cj3i1);
         C[1+i2] -:= gkl*(ci2j1 + cj1i2 +
			  ci2j3 + cj3i2);
         C[1+i3] -:= gkl*(ci3j1 + cj1i3 +
                          ci3j2 + cj2i3);

	 C[1+j1] +:= gkl*(cj1i2 + ci2j1 +
	                  cj1i3 + ci3j1);
	 C[1+j2] +:= gkl*(cj2i1 + ci1j2 +
	                  cj2i3 + ci3j2);
         C[1+j3] +:= gkl*(cj3i1 + ci1j3 +
                          cj3i2 + ci2j3);

       end for;
   end for;
   
   /* Case k = l */
   for k := 1 to d-1  do
       c :=  A[1+k]*B[1+k];

       C[1+0] +:= c*(EllBasis`Xs0[1+k]-Xi[1+k]);

       for l := 1 to k-1 do
	   C[1+l] +:= c*Xs[1+((d-k+l) mod d)];
       end for;

       C[1+k] -:= c*XsS;

       for l := k+1 to d-1 do
	   C[1+l] +:= c*Xs[1+((d-k+l) mod d)];
       end for;
   end for;

   return C;
end intrinsic;

/* Elliptic Basis Multiplication 

   If possible, programmed in a ``quasi-linear complexity in d'' style.
*/
intrinsic EllNormalBasisMultiply(A::ModTupFldElt,B::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
    {Multiplication of two elements given in an elliptic normal basis}

    if EllBasis`FastAlgorithmsAvailable eq true then
	return  EllNormalBasisFastMultiply(A, B, EllBasis);
    end if;

    EA := EllNormalBasisToBasis(A, EllBasis);
    EB := EllNormalBasisToBasis(B, EllBasis);

    EC := EllBasisMultiply(EA, EB, EllBasis);

    C  := EllBasisToNormalBasis(EC, EllBasis);

    return C;
end intrinsic;

/* Elliptic Basis Inversion with Lagrange's theorem 
intrinsic EllBasisInverse(A::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
    {Inverse of an element given in the elliptic basis}

    Fq := EllBasis`Fq; d := EllBasis`d;
    q := #Fq;
    
    // First, Y = A ^ (q-2)
    exp := q - 2; T := A;
    if (exp mod 2) eq 1 then
	Y := A;
    else
	Y := VectorSpace(Fq, d)!0; Y[1] := 1;
    end if;

    exp := exp div 2;
    while exp ne 0 do
	T := EllBasisMultiply(T, T, EllBasis);
	if (exp mod 2) eq 1 then
	    Y := EllBasisMultiply(Y, T, EllBasis);
	end if;
	exp := exp div 2;
    end while;

    // Z = A * Y 
    Z := EllBasisMultiply(Y, A, EllBasis);

    // Z ^ ((q^(d-1) - 1) / (q - 1))
    T := Z;
    for k := 1 to d-2 do
	T := EllBasisFrobenius(T, EllBasis);
	Z := EllBasisMultiply(T, Z, EllBasis);
    end for;

    // A ^ (q^d-2)
    Z := EllBasisFrobenius(Z, EllBasis);
    Z := EllBasisMultiply(Y, Z, EllBasis);
    
    return Z;
end intrinsic;
*/

/* Elliptic Basis Inversion
 */
intrinsic EllBasisInverse(A::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
    {Inverse of an element given in the elliptic basis}

    a := EllBasisToPolynomial(A, EllBasis);
    return EllPolynomialToBasis(1/a, EllBasis);
    
end intrinsic;

/* Elliptic Normal Basis Inversion

   A very classical binary exponentiation
 */
intrinsic EllNormalBasisInverse(A::ModTupFldElt,EllBasis::Rec) -> ModTupFldElt
    {Inverse of an element given in the elliptic basis}

    EA := EllNormalBasisToBasis(A, EllBasis);
    EC := EllBasisInverse(EA, EllBasis);
    C  := EllBasisToNormalBasis(EC, EllBasis);

    return C;
end intrinsic;

/* Elliptic Basis Exponentiation

   A very classical binary exponentitation
 */
intrinsic EllBasisExp(A::ModTupFldElt,e::RngIntElt,EllBasis::Rec) -> ModTupFldElt
    {d-th power of an element given in an elliptic basis}

    exp := e; T := A;
    if (exp mod 2) eq 1 then
	C := A;
    else
	C := VectorSpace(EllBasis`Fq, EllBasis`d)!0; C[1] := 1;
    end if;
    exp := exp div 2;
    while exp ne 0 do
	T := EllBasisMultiply(T, T, EllBasis);
	if (exp mod 2) eq 1 then
	    C := EllBasisMultiply(C, T, EllBasis);
	end if;
	exp := exp div 2;
    end while;

    return C;
end intrinsic;

/* Elliptic Normal Basis Exponentiation

   A very classical binary exponentiation
 */
intrinsic EllNormalBasisExp(A::ModTupFldElt,e::RngIntElt,EllBasis::Rec) -> ModTupFldElt
    {d-th power of an element given in an elliptic normal basis}

    exp := e; T := A;
    if (exp mod 2) eq 1 then
	C := A;
    else
	C := VectorSpace(EllBasis`Fq, EllBasis`d)![1 : i in [1..EllBasis`d]]; 
    end if;
    exp := exp div 2;
    while exp ne 0 do
	T := EllNormalBasisMultiply(T, T, EllBasis);
	if (exp mod 2) eq 1 then
	    C := EllNormalBasisMultiply(C, T, EllBasis);
	end if;
	exp := exp div 2;
    end while;

    return C;
end intrinsic;

/* Initialising elliptic basis

   This function precomputes whatever is needed to handle elliptic (normal)
   basis.
*/
intrinsic EllBasisCreate(Fq::FldFin, d::RngIntElt :
    CheckIfEllipticBasisExist := true, FastAlgorithmsOnly := false) -> Rec
    {Generate an elliptic (normal) basis of a degree d extension over Fq}

    /* Easy initialisations
     **********************/
    p := Characteristic(Fq); q := #Fq;
    Px := PolynomialRing(Fq); x := Px.1;
    found := false; naive := false;
    nbtries := 0; nbloop := 0;
    
    /* Do we have reasonable hope to find such an elliptic basis ?
     *************************************************************/
    if CheckIfEllipticBasisExist and not EllIsPossible(Fq, d) then

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] Bad degree d, I give up\n";
end if;

	return rec<EllBasis | Initialised := false>;
    end if;

    /* We have to find an elliptic curve with good properties,
     * let's go !
     ***********************************************************/
    repeat
	nbtries +:= 1; if nbtries ge 10 then nbtries := 1; end if;
	
	if nbtries eq 1 then
	    
	    /* Getting an  Elliptic Curve  */
	    Ec, KP := EllFindEcOfPrescribedOrder(Fq, d);
	    a1, a2, a3, a4, a6 := Explode(aInvariants(Ec));

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] %o with j-invariant %o and order %o found.\n",
	Ec, jInvariant(Ec), Order(Ec);
    printf "[EllBasis] With point %o is of order %o.\n", KP, d;
end if;

            /* And the endomorphism of prescribed Kernel */
	    Q := Ec!KP; TD := [[x-Q[1]]];
	    for i := 2 to (d div 2) do
		Q +:= KP; Append(~(TD[1]), x-Q[1]);
	    end for;
	    EllEvaluationTree(~TD); DX := TD[#TD][1];
	    Ep, In := IsogenyFromKernel(Ec, DX : Check := false);	
	    nx := IsogenyMapPhi(In); dx := IsogenyMapPsiSquared(In);
	    Tmp := Gcd(nx, dx); nx := nx / Tmp; dx := dx / Tmp;

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] %o is then isogenous of degree %o to this curve.\n", Ep, d;
end if;

        end if;

	/* The extension of degree d */
	repeat pt := Random(Ep); until pt ne Ep!0;
	eqP := Px!(nx-pt[1]*dx);

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] We now try the point %o on the isogenous curve.\n", pt;
end if;

        if not IsIrreducible(eqP) then

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] ... but it does not yield an irreducible pre-image.\n";
end if;

            continue;
	end if;

	Fqd := ext<Fq|eqP/Coefficient(eqP, Degree(eqP))>; t := Fqd.1;

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] It works, the degree %o extension is defined by %o.\n",
	d, DefiningPolynomial(Fqd);
end if;

        /* Is the generic point of good order ? */
	En :=  BaseExtend(Ec, Fqd);
	P := Points(En, t)[1]; xP, yP, _ := Explode(Eltseq(P));
	if d*P eq En!0 then
	    
if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] But the generic point is of order %o too, I give up this try.\n", d;
end if;

            continue;
	end if;
		
	/* The generator of the Galois action */
	LP := Ec!(En![x^q : x in Eltseq(P)]-P);

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] The point %o generates the kernel of the degree %o isogeny.\n", LP, d;
end if;
	
	/* Multiplication data */
	Q := Ec!0; Xi := [Q[1]]; Yi := [Q[2]];
	for i := 1 to (d - 1) do
	    Q +:= LP; Append(~Xi, Q[1]); Append(~Yi, Q[2]);
	end for;

	/* Fast conversion data */
	py1, py2 := EllConvergent(eqP, Px!Eltseq(yP, Fq), d div 2);
	pg := Gcd(py1, py2); py1 := py1 div pg; py2 :=  py2 div pg;
	if Degree(py1) eq (d div 2) and Degree(py2) eq ((d-1) div 2) then
	    cst := 1;
	    if IsEven(d) then
		xi := Xi[1+(d div 2)]; yi := Yi[1+(d div 2)];
		cst := -Evaluate(py1, xi)+yi*Evaluate(py2, xi);
	    end if;
	    if cst ne 0 then
		_, pi1, pi2 := ExtendedGreatestCommonDivisor(py1, py2);
		found := true;
	    end if;
	end if;

	nbloop +:= 1;
	if nbloop ge 10 then

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] Impossible to write the ordinate yP of the generic point	as y1/y2\n";
    printf "[EllBasis] with y1 and y2 of degree d/2 in the degree %o extension.\n", d;
    printf "[EllBasis] Fast Polynomial to Elliptic Basis is not possible with our implementation.\n";
end if;

            if FastAlgorithmsOnly then

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] Only fast algorithms are required, so I give up.\n";
end if;

	      return rec<EllBasis | Initialised := false>;
	    end if;

	    found := true; naive := true;
	end if;

	/* Everything's fine ? */
	if not found then

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] But it does not work.\n";
end if;

        end if;

    until found eq true;

    /* At this point, we found an appropriate curve
     ************************************************/
    
    /* agot, bgot, cgot */
    if p eq 2 or p eq 3 then Fq2 := GF(#Fq^4); else Fq2 := GF(#Fq^2); end if;
    Ec2 := BaseExtend(Ec, Fq2);
    repeat Pp := Random(Ec2); until Evaluate(DX, Pp[1]) ne 0;
    Qp := Ec!0; Q := Ec!LP; cstc := Fq2!EllUAB(Qp, Q, Pp, Ec);
    for i := 1 to d-1 do
	Qp := Q; Q +:= LP; cstc +:= Fq2!EllUAB(Qp, Q, Pp, Ec);
    end for;
    cstc := Fq!cstc;
    if cstc eq 0 then
	csta := Fq ! 1; cstb := (Fq ! 1) / d; 		/* d can not be equal to 0 mod p
                                                           in this case */
    else
	csta := 1 / cstc; cstb := Fq ! 0;
    end if;

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] We have cgot = %o, agot = %o and bgot = %o.\n", cstc, csta, cstb;
end if;

    /* Initialising the structure */
    EllB := rec<EllBasis |
	Initialised := true, FastAlgorithmsOnly := FastAlgorithmsOnly,
	Fq := Fq, d := d, Fqd :=  Fqd,
	Ec := Ec, Xi := Xi, Yi := Yi,
	csta := csta, cstb := cstb,
	NaiveConversion := naive
	>;

    /* Polynomial <-> Elliptic Basis conversion precomputations
     ***********************************************************/
    if naive eq true then

	/* Change of variable between the polynomial basis and the
          elliptic basis */
	B := [Fqd!1];
	for i := 1 to d-1 do
	    xi := EllB`Xi[1+i]; yi := EllB`Yi[1+i];
	    B cat:= [(yP + yi + a1 * xi + a3)/(xP - xi)];
	end for;
	EllB`B := B;
	Mat := RMatrixSpace(Fq, d, d)!&cat[Eltseq(x, Fq) : x in B];
	EllB`MB := Mat^(-1);

	/* The normal elliptic basis */
	Qp := Ec!0; Q := Ec!LP; NB := [Fqd!EllUAB(Qp, Q, P, Ec)];
	for i := 1 to d-1 do
	    Qp := Q; Q +:= LP; Append(~NB, EllUAB(Qp, Q, P, Ec));
	end for;
	for i := 1 to d do NB[i] := csta * NB[i] + cstb; end for;
	EllB`NB := NB;
	Mat := RMatrixSpace(Fq, d, d)!&cat[Eltseq(x, Fq) : x in NB];
	EllB`MNB := Mat^(-1);

    else

	/* The tree of TT (X-xi) */
	TD := [[x-Xi[1+i] : i in [1..(d div 2)]]];
	EllEvaluationTree(~TD);

	/* Evaluation of diff(DX) at Xi's */
	LD := [Derivative(DX) mod DX];
	for lvl := #TD-1 to 1 by -1 do
	    NewLD := []; ptr := 1;
	    EllFastEvaluate(~NewLD, ~ptr, LD, TD[lvl], 1, #TD[lvl]+1);
	    LD := NewLD;
	end for;

	/* Additional evaluation point when d is even */
	if IsEven(d) then
	    xi := EllB`Xi[2];
	    for _xi in Fq do
		yb := -a1*xi-a3;
		if Evaluate(DX, _xi) ne 0 and
		    (yb + EllB`Yi[1+(d div 2)] + a1 * EllB`Xi[1+(d div 2)] + a3) ne 0 then
		    xi := _xi; break;
		end if;
	    end for;
	    if xi eq EllB`Xi[2] then
		for _xi in Fq2 do
		    Pts := Points(Ec2, _xi); if Pts eq [] then continue; end if;
		    yb := Pts[1][2];
		    if Evaluate(DX, _xi) ne 0 and
			(yb + EllB`Yi[1+(d div 2)] + a1 * EllB`Xi[1+(d div 2)] + a3) ne 0 then
			xi := _xi; break;
		    end if;
		end for;
	    end if;
	    EllB`xi := xi; EllB`yb := yb;
	end if;
	EllB`py1 := py1; EllB`py2 := py2; EllB`pi1 := pi1; EllB`pi2 := pi2;
	EllB`dn := Fqd!Eltseq(DX); EllB`yP := yP; EllB`TD := TD; EllB`LD := LD;

    end if;

    /* Elliptic Basis Multiplication Precomputations
     ***********************************************/
    if not FastAlgorithmsOnly then
	Xs := EllPolynomialToBasis(xP, EllB);
	EllB`Xs := Xs;
	EllB`XsS := &+( Eltseq(Xs)[2..d]);
    
	EllB`Xs0 := [];   
	Xs0 := [Xs[1]];
	for k := 1 to d-1 do
	    Xs :=  EllBasisFrobeniusInverse(Xs, EllB);
	    Xs0 cat:= [Xs[1]];
	end for;
	EllB`Xs0 := Xs0;
    end if;
	
    /* Normal Basis Fast Multiplication Precomputations
     ***************************************************/
    Gp, Mp := AbelianGroup(EllB`Ec); R := EllB`Ec!0;
    for g in Setseq(Generators(Gp)) do
	if d mod Order(Mp(g)) ne 0 then R := Mp(g); break; end if;
    end for;
    
    if R eq EllB`Ec!0 then
	EllB`FastAlgorithmsAvailable := false;

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] Impossible to find a point on the curve out of the kernel of the degree %o isogeny.\n", d;
    printf "[EllBasis] Our fast normal basis multiplication algorithm is not available.\n";
end if;

        if FastAlgorithmsOnly then

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] Only fast algorithms are required, so I give up.\n";
end if;

            return rec<EllBasis | Initialised := false>;
	end if;

    else
	Xr := VectorSpace(Fq, d)!0; Ur := VectorSpace(Fq, d)!0;
	Q := R; Xr[1] := Eltseq(Q)[1]; Ur[1] := csta * EllUOA(LP, Q, Ec) + cstb;
	for i := 1 to d-1 do
	    Q +:= LP; Xr[1+i] := Eltseq(Q)[1]; Ur[1+i] := csta * EllUOA(LP, Q, Ec) + cstb;
	end for;
	Xr := Xr * csta^2;
	Iota := EllPolynomialToNormalBasis(Fqd!t, EllB); Iota := Iota * csta^2;
	_, ax, _ := ExtendedGreatestCommonDivisor(Px!Eltseq(Ur), x^d-1);
	AX := Eltseq(ax); AX cat:= [Fq!0 : i in [#AX+1..d]];
	Uri := VectorSpace(Fq, d)!AX;

	EllB`Ur := Ur; EllB`Uri := Uri; EllB`Xr := Xr; EllB`Iota := Iota;
	EllB`FastAlgorithmsAvailable := true;
    end if;

if GetVerbose("EllBasis") ge 1 then
    printf "[EllBasis] Elliptic (Normal) Basis succesfully initialised.\n";
end if;

    return EllB;

end intrinsic;
